#ifndef HASH_TABLE_H_INCLUDED
#define HASH_TABLE_H_INCLUDED

/*
It is assumed that there is less than UINT_MAX * sizeof(void*)
bytes of memory available, or that one will never try to set
UINT_MAX * 3/4 different keys.  If this is not true then integer
overflow may occur and things will misbehave quite well.
*/

#define HASH unsigned long

struct hash_table { char dummy[1]; }; /* type is opaque outside of implementation file */ 

#ifdef __cplusplus
extern "C"
{
#endif

/*
  Creates a hash table of the given size.
    nSize must be a power of 2.
    will return NULL if allocation fails.
*/
struct hash_table *HashTable_Create(
    unsigned nSize);

/*
  Destroys a hash table
    pTable may be NULL.
    will return NULL as a convenience.  (always)
    all entries and their keys/values will also be freed.
*/
struct hash_table *HashTable_Destroy(
    struct hash_table *pTable);

/*
  Associates a given hash/key with a value.
    pTable may be NULL.
    key and value must not be NULL.
    will only return NULL if pTable is NULL and allocation fails.
    may return a different value than pTable, in which case
      pTable is no longer valid.
    stores copies of key and value in the table, not just
      pointers.
    key matches are case insensitive.

  The hash table will automatically grow when entries are added
  beyond a certain density, unless there is not enough memory
  to do so.
*/
struct hash_table *HashTable_Set(
    struct hash_table *pTable,
    HASH hash,
    const char *key,
    const char *value);

/*
  Retrieves the value associated with a given hash/key.
    pTable may be NULL.
    key must not be NULL.
    will return NULL if the key is not in the table or if
      pTable is NULL.
    key matches are case insensitive.
*/
const char *HashTable_Get(
    const struct hash_table *pTable,
    HASH hash,
    const char *key);

#ifndef NLOG
void HashTable_Dump(const struct hash_table *pTable);
#endif

#ifdef __cplusplus
}
#endif

#endif /* HASH_TABLE_H_INCLUDED */ 
