/******************************************************************************
*                                                                             *
* This is a part of the LsxCommand LiteStep module Source code.               *
*                                                                             *
* Copyright (C) 1999-2000 Visigoth (Shaheen Gandhi in real life)              *
* Based on limpid's lscommand.                                                *
* Updated by blkhawk, ilmcuts, jesus_mjjg, rabidcow.                          *
* Look at the documentation for more informations.                            *
*                                                                             *
* This program is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by        *
* the Free Software Foundation; either version 2 of the License, or           *
* (at your option) any later version.                                         *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA   *
*                                                                             *
******************************************************************************/

/******************************************************************************
*                lsxcommand.c - Main lsxcommand source file                   *
*                                   *  *  *  *                                *
*           Last Update:  October 31, 2003  10:20 A M                         *
******************************************************************************/

#include <windows.h>
#include <shlwapi.h>
#include <shlobj.h>
#include <stdio.h>
#include <time.h>
#include <locale.h>
#include "resource.h"
#include "exports.h"
#include "lsxcommand.h"
#include "hash_table.h"
#include "list.h"

#include "ModuleLog.h"

/******************************************************************************
*	MSVC++ 6.0                                                                  *
*	Go to Tools -> Options -> Directories                                       *
*	Select "Include file" in the list box and add the path to the directory     *
*	where you store your litestep core source files.                            *
*                                                                             *
*	Don't forget to do the same for the library file.                           *
*                                                                             *
*	Please do this to help portability and to support some development          *
*   standards.                                                                *
******************************************************************************/
#include <lsapi/lsapi.h>

//20021115
int msgs[] = {LM_GETREVID, 0};
const char szAppName[] = "lsxcommand.dll"; // Our window class, etc
const char rcsRevision[] = "1.9.1"; // Our Version

//20021114
#define MAX_LINE_LENGTH 4096

#ifndef LSXCOMMANDCLOCK_EXPORTS

BOOL editfirst = FALSE, staticfirst = FALSE;
char *szApp = "LSXCommand";
int nHistoryEntries = 0, nFiles = 0;
HMENU hContextMenu = NULL;
MENUITEMINFO item;

extern double Evaluate(char *expr, int *error, char *replacement, size_t replace_size, struct CommandSettings *cs);

#else

char *szApp = "LSXCommandClock";

#endif //LSXCOMMANDCLOCK_EXPORTS

BOOL lsboxed = FALSE;
char szModuleIniPath[MAX_PATH];
enum Timers curTimer = NO_TIMER;
HINSTANCE hInst = NULL;
HWND hWnd = NULL, hText = NULL, hBG = NULL;
HFONT hFont = NULL;
HBRUSH hBGBrush = NULL, hHollowBrush = NULL;
HBITMAP hBGBitmap = NULL;
WNDPROC wpOld, wpBG;
struct CommandSettings *cs;

#ifdef _MSC_VER
/* inline is C++ only, MSVC allows __inline in C to do the same thing */ 
/* (not to be confused with __forceinline)                            */ 
#define INLINE __inline
#else
/* as for other compilers, I don't know.                              */ 
#define INLINE
#endif

/***********************************************************
* Short, commonly used commands.                           *
***********************************************************/
INLINE void TriggerEvent(unsigned event) { PostMessage(hWnd, CM_TRIGGEREVENT, event, 0); }
INLINE void SelectAll(void) { SendMessage(hText, EM_SETSEL, 0, -1); }
INLINE void ClearText(void) { SetWindowText(hText,""); }
INLINE void SetFocusGeneric(void) /* may fail, that's life */ 
{
  if (!IsWindowVisible(hWnd))
    ShowWindow(hWnd, SW_SHOW);
  SetForegroundWindow(hText);
  SetFocus(hText);
}
INLINE void SetFocusKeyboard(void) { SetFocusGeneric(); if (cs->SelectAllOnFocus)      SelectAll(); }
INLINE void SetFocusMouse(void)    { SetFocusGeneric(); if (cs->SelectAllOnMouseFocus) SelectAll(); }
INLINE void SetCursorAfterText(const char *sText)
{
  const unsigned nText = strlen(sText);
  SetWindowText(hText, sText);
  SendMessage(hText, EM_SETSEL, nText, nText);
}
INLINE void SetFocusWithText(const char *sText) { SetFocusGeneric(); SetCursorAfterText(sText); }
INLINE void SetSelectedText(const char *sText) { SetWindowText(hText, sText); SelectAll(); }
INLINE void SetFocusWithTextSelected(const char *sText) { SetFocusGeneric(); SetSelectedText(sText); }

/***********************************************************
* struct CommandSettings *ReadSettings()                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - wharfDataType *wd                                   *
*      Pointer to Litestep's structure for general info.   *
*                         *  *  *  *                       *
* Retrieves settings in the RC file and stores it all in a *
* local structure. Returns a pointer to a new structure.   *
***********************************************************/

struct CommandSettings *ReadSettings(LPCSTR lsPath)
{
  int offsetx, offsety;
  RECT screen;
  char szTemp[MAX_LINE_LENGTH];

  struct CommandSettings *settings = malloc(sizeof(*settings));

  settings->OnFocusCommand = NULL;
  settings->OnUnfocusCommand = NULL;

#ifndef LSXCOMMANDCLOCK_EXPORTS
#define RCSETTING(STRING) "Command" STRING
#else
#define RCSETTING(STRING) "CommandClock" STRING
#endif

  if (!GetRCString(RCSETTING("HistoryFile"), szModuleIniPath, NULL, MAX_PATH))
  {
      strcpy(szModuleIniPath, lsPath);
      strcat(szModuleIniPath, "\\MODULES.INI");
  }
  
  GetRCString(RCSETTING("TextFontFace"),settings->TextFontFace,"Arial",256);
  settings->BGColor = GetRCColor(RCSETTING("BGColor"),RGB(255,255,255));
  settings->TextColor = GetRCColor(RCSETTING("TextColor"),RGB(0,0,0));
  settings->TextSize = GetRCInt(RCSETTING("TextSize"),14);

  settings->m_bTextBold = settings->m_bTextUnderline = settings->m_bTextItalic = FALSE;
  settings->m_bTextBold = GetRCBool(RCSETTING("TextBold"),TRUE);
  settings->m_bTextUnderline = GetRCBool(RCSETTING("TextUnderline"),TRUE);
  settings->m_bTextItalic = GetRCBool(RCSETTING("TextItalic"),TRUE);

  if (GetRCString(RCSETTING("TextAlign"), szTemp, NULL, MAX_LINE_LENGTH))
  {
    if (stricmp(szTemp, "left") == 0)
      settings->TextAlign = ES_LEFT;
    else if (stricmp(szTemp, "center") == 0)
      settings->TextAlign = ES_CENTER;
    else if (stricmp(szTemp, "right") == 0)
      settings->TextAlign = ES_RIGHT;
  }
  else
    settings->TextAlign = ES_LEFT;

  /* primary monitor */ 
  screen.left = 0;
  screen.right = GetSystemMetrics(SM_CXSCREEN);
  screen.top = 0;
  screen.bottom = GetSystemMetrics(SM_CYSCREEN);

  settings->origx = (settings->x = GetRCCoordinate(RCSETTING("X"), 0, screen.right-screen.left)) + screen.left;
  settings->origy = (settings->y = GetRCCoordinate(RCSETTING("Y"), 0, screen.bottom-screen.top)) + screen.top;
  settings->origwidth = settings->width = GetRCCoordinate(RCSETTING("Width"), 160, screen.right-screen.left);
  settings->origheight = settings->height = GetRCCoordinate(RCSETTING("Height"), 20, screen.bottom-screen.top);
  settings->BorderSize = GetRCInt(RCSETTING("BorderSize"), -1);
  offsetx = GetRCInt(RCSETTING("OffsetX"), 0);
  offsety = GetRCInt(RCSETTING("OffsetY"), 0);

  settings->BorderColor = GetRCColor(RCSETTING("BorderColor"),RGB(102,102,102));
  settings->BevelBorder = GetRCBool(RCSETTING("BevelBorder"),TRUE);
  settings->NoAlwaysOnTop = GetRCBool(RCSETTING("NotAlwaysOnTop"),TRUE);
  settings->notmoveable = GetRCBool(RCSETTING("NotMoveable"),TRUE);
  settings->HiddenOnStart = GetRCBool(RCSETTING("HiddenOnStart"),TRUE);
  settings->WaitForBox = GetRCBool(RCSETTING("WaitForBox"),TRUE);
  settings->NoCursorChange = GetRCBool(RCSETTING("NoCursorChange"),TRUE);
  settings->SelectAllOnFocus = GetRCBool(RCSETTING("SelectAllOnFocus"), TRUE);
  settings->SelectAllOnMouseFocus = GetRCBool(RCSETTING("SelectAllOnMouseFocus"), TRUE);
  settings->Transparent = GetRCBool(RCSETTING("TransparentEditBox"), TRUE);
  settings->ScrollWinAmp = GetRCBool(RCSETTING("ScrollWinAmpTitle"), TRUE);
  settings->HideOnUnfocus = GetRCBool(RCSETTING("HideOnUnfocus"), TRUE);
  GetRCString(RCSETTING("Background"), settings->Background, "", sizeof(settings->Background));
  settings->m_bStretchImage = GetRCBoolDef(RCSETTING("StretchBackground"), FALSE);

#ifndef LSXCOMMANDCLOCK_EXPORTS
  /* settings not used by commandclock */ 
  settings->MaxHistoryEntries = GetRCInt(RCSETTING("HistoryEntries"), 10) + 1;
  settings->MaxHistoryMenuEntries = GetRCInt(RCSETTING("HistoryMenuEntries"), settings->MaxHistoryEntries - 1 );
  settings->ContextMenuStandardItems = GetRCInt(RCSETTING("ContextMenuStandardItems"), 0);
#define STANDARDMENUITEMS_ABOVE 1
#define STANDARDMENUITEMS_BELOW 2
  settings->NoClearOnCommand = GetRCBool(RCSETTING("NoClearOnCommand"),TRUE);
  settings->HideOnCommand = GetRCBool(RCSETTING("HideOnCommand"),TRUE);
	settings->HideOnEscape = GetRCBool(RCSETTING("HideOnEsc"),TRUE);
  settings->ClearOnHide = GetRCBool(RCSETTING("ClearOnHide"),TRUE);
  settings->AssumeNetAddress = GetRCBool(RCSETTING("AssumeNetAddress"), TRUE);
  settings->NoWarnOnError = GetRCBool(RCSETTING("NoWarnOnError"), TRUE);
  settings->NoTabMicroComplete = GetRCBool(RCSETTING("NoTabMicroComplete"), TRUE);
  settings->TabFileComplete = GetRCBool(RCSETTING("TabFileComplete"), TRUE);
  settings->ClearHistoryOnStartup = GetRCBool(RCSETTING("ClearHistoryOnStartup"), TRUE);
  settings->WrapHistory = GetRCBool(RCSETTING("WrapHistory"), TRUE);
  settings->NewestHistoryItemsOnTop = GetRCBool(RCSETTING("NewestHistoryItemsOnTop"), TRUE);
  settings->ContextMenuAutoPopup = GetRCBool(RCSETTING("ContextMenuAutoPopup"), TRUE);
  settings->ContextMenuAboveBox = GetRCBool(RCSETTING("ContextMenuAboveBox"), TRUE);
  settings->ContextMenuExecute = GetRCBool(RCSETTING("ContextMenuExecute"), TRUE);
  settings->UnixHistory = GetRCBool(RCSETTING("UnixHistory"), TRUE);
  settings->ExplorePaths = GetRCBool(RCSETTING("ExplorePaths"), TRUE);
  settings->AutoComplete = GetRCBool(RCSETTING("NoAutoComplete"), FALSE);
  settings->CommaDelimiter = GetRCBool(RCSETTING("CommaDelimiter"), TRUE);
  settings->RPNCalculator = GetRCBool(RCSETTING("RPNCalculator"), TRUE);
  settings->ClockDisappears = GetRCBool(RCSETTING("ClockDisappearsOnFocus"), TRUE);
  settings->WinAmpDisappears = GetRCBool(RCSETTING("WinAmpDisappearsOnFocus"), TRUE);
  settings->AddExternalsToHistory = GetRCBool(RCSETTING("AddExternalsToHistory"), TRUE);
  GetRCString(RCSETTING("SearchEngineList"), settings->SearchEngineList, "engines.list", sizeof(settings->SearchEngineList));
  GetRCString(RCSETTING("ContextMenuOrder"), settings->ContextMenuOrder, "012", sizeof(settings->ContextMenuOrder));
  GetRCString(RCSETTING("Time"), settings->Clock, "", sizeof(settings->Clock));
  GetRCString(RCSETTING("SearchEngineBrowser"), settings->BrowserPath, "", sizeof(settings->BrowserPath));
  GetRCString(RCSETTING("MusicPlayer"), settings->MusicPlayer, "WINAMP V1.X", sizeof(settings->MusicPlayer));
#else
  /* settings specific to commandclock */ 
  GetRCString("CommandClockString", settings->Clock, "", sizeof(settings->Clock));
#endif
  
  /* Special Cases */ 

  /* If there's a bevel, then no border                       */ 
  if (settings->BevelBorder) settings->BorderSize = 0;

  /* If the universal border is invalid, use separate borders */ 
	if (settings->BorderSize < 0)
  {
    settings->BottomBorderSize = GetRCInt(RCSETTING("BottomBorderSize"),2);
    settings->TopBorderSize = GetRCInt(RCSETTING("TopBorderSize"),2);
    settings->LeftBorderSize = GetRCInt(RCSETTING("LeftBorderSize"),2);
    settings->RightBorderSize = GetRCInt(RCSETTING("RightBorderSize"),2);
	} else {
    settings->BottomBorderSize = settings->BorderSize;
    settings->TopBorderSize = settings->BorderSize;
    settings->LeftBorderSize = settings->BorderSize;
    settings->RightBorderSize = settings->BorderSize;
	}

  /* No Background */ 
  if(!(*(settings->Background)))
    settings->Transparent = 0;

  /* Offsets */ 
  /*
#define OFFSET_LEFT 0
#define OFFSET_CENTER 1
#define OFFSET_RIGHT 2
  switch (offsetx)
  {
  case OFFSET_LEFT:
    if(settings->x < 0)
      settings->origx = settings->x = screen.right + settings->x;
    break;
  case OFFSET_CENTER:
    settings->origx = settings->x =
      (screen.right+screen.left) / 2 - settings->width / 2 + settings->x;
    break;
  default:
    settings->origx = settings->x =
      screen.right + settings->x;
  }

  if(offsety == 0) {
    if(settings->y < 0)
      settings->origy = settings->y = screen.bottom + settings->y;
  } else if(offsety == 1)
    settings->origy = settings->y = screen.bottom / 2 - settings->height / 2 + settings->y;
  else
    settings->origy = settings->y = screen.bottom + settings->y;
    */

	GetRCLine(RCSETTING("OnFocus"), szTemp, MAX_LINE_LENGTH, "!NONE");
	settings->OnFocusCommand = malloc((strlen(szTemp)+1)*sizeof(*settings->OnFocusCommand));
	strcpy(settings->OnFocusCommand, szTemp);
  
	GetRCLine(RCSETTING("OnUnfocus"), szTemp, MAX_LINE_LENGTH, "!NONE");
	settings->OnUnfocusCommand = malloc((strlen(szTemp)+1)*sizeof(*settings->OnUnfocusCommand));
	strcpy(settings->OnUnfocusCommand, szTemp);

  ReadCommandChars(settings);
#undef RCSETTING
	return settings;
}

/***********************************************************
* void ContextMenuInit()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*                         *  *  *  *                       *
* Initializes the context menu.  This is called whenever   *
* one of the submenu handles is changed.                   *
***********************************************************/

#ifndef LSXCOMMANDCLOCK_EXPORTS
void ContextMenuInit()
{
/* This is so wrong I just can't bring myself to delete it.

  char buf[15], *p = cs->ContextMenuOrder, atom;
  int i, indices[7];

  for(i = 0; i < 3; ++i)
  {
    atom = *p;
    indices[i] = atoi(&atom);
    ++p;
  }
*/

  const char *pMenuOrder = cs->ContextMenuOrder;
  UINT Flags;
  int iInsert;
  unsigned nNines;

  TRACE1("ContextMenuInit");

  if (hContextMenu)
    DestroyMenu(hContextMenu);

  hContextMenu = CreatePopupMenu();
  AppendMenu(hContextMenu, MF_STRING|MF_GRAYED, 0, "(check settings)");
  AppendMenu(hContextMenu, MF_STRING|MF_GRAYED, 1, "(check settings)");
  AppendMenu(hContextMenu, MF_STRING|MF_GRAYED, 2, "(check settings)");

  nNines = 0;

  /* history menu */ 
  if (*pMenuOrder >= '0' && *pMenuOrder <= '2')
  {
    Flags = MF_STRING | (hHistoryMenu ? MF_POPUP|MF_ENABLED : MF_GRAYED);
    ModifyMenu(hContextMenu, (*pMenuOrder-'0'), Flags, (UINT)hHistoryMenu, "&History");
  }
  else if (*pMenuOrder == '9')
    ++nNines;
  if (*pMenuOrder) ++pMenuOrder;

  /* alias menu */ 
  if (*pMenuOrder >= '0' && *pMenuOrder <= '2')
  {
    Flags = MF_STRING | (hAliasMenu ? MF_POPUP|MF_ENABLED : MF_GRAYED);
    ModifyMenu(hContextMenu, (*pMenuOrder-'0'), Flags, (UINT)hAliasMenu, "&Aliases");
  }
  else if (*pMenuOrder == '9')
    ++nNines;
  if (*pMenuOrder) ++pMenuOrder;

  /* search engine menu */ 
  if (*pMenuOrder >= '0' && *pMenuOrder <= '2')
  {
    Flags = MF_STRING | (hSearchMenu ? MF_POPUP|MF_ENABLED : MF_GRAYED);
    ModifyMenu(hContextMenu, (*pMenuOrder-'0'), Flags, (UINT)hSearchMenu, "&Search Engines");
  }
  else if (*pMenuOrder == '9')
    ++nNines;
  if (*pMenuOrder) ++pMenuOrder;

  /* remove invalid items if anything was set to 9 */ 
  /* this is not strictly correct, but only fails by improving */ 
  if (nNines)
    for (iInsert = 0; iInsert < 3; )
    {
      if (GetMenuState(hContextMenu, iInsert, MF_BYPOSITION) & MF_GRAYED)
        RemoveMenu(hContextMenu, iInsert, MF_BYPOSITION);
      else
        ++iInsert;
    }

  /* "standard" items */ 
  switch (cs->ContextMenuStandardItems)
  {
  case STANDARDMENUITEMS_ABOVE:
    iInsert =  0;
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_SEPARATOR, 0, NULL);
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_STRING|MF_ENABLED, HMI_USER_PASTE, "&Paste");
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_STRING|MF_ENABLED, HMI_USER_COPY, "&Copy");
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_STRING|MF_ENABLED, HMI_USER_CUT, "Cu&t");
    break;

  case STANDARDMENUITEMS_BELOW:
    iInsert = -1;
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_SEPARATOR, 0, NULL);
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_STRING|MF_ENABLED, HMI_USER_CUT, "Cu&t");
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_STRING|MF_ENABLED, HMI_USER_COPY, "&Copy");
    InsertMenu(hContextMenu, iInsert, MF_BYPOSITION|MF_STRING|MF_ENABLED, HMI_USER_PASTE, "&Paste");
    break;

  default: return;
  }
}

/***********************************************************
* void ParseCalculatorCommand()                            *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - char *command                                       *
*      Search Engine string.                               *
*                         *  *  *  *                       *
* This function is called when the user wants to calculate *
* a value using the calculator.                            *
***********************************************************/

void ParseCalculatorCommand(char *command)
{
  int error = 0;
  char *p = NULL, val[1024];

  ++command;
  if(cs->CommaDelimiter) {
    p = command;
    while(*p) {
      if(*p == ',')
        *p = '.';
      ++p;
    }
  }

  Evaluate(command, &error, val, sizeof(val), cs);

  if(!error || error == ERROR_CONVERSIONS_DONE)
  {
    if(cs->CommaDelimiter)
    {
      for (p=val; *p; ++p)
        if (*p == '.')
          *p = ',';
    }
    SetFocusWithText(val);
  }
  else
  {
    SetFocusWithTextSelected("error");
  }
}

enum GuessType
{
  eGuessDunno,
  eGuessBang,
  eGuessFolder,
  eGuessFile,
  eGuessURL,
  eGuessURLNoProtocol
};

/***********************************************************
* enum GuessType GuessAnalysis()                           *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *sCommand;                               *
*      The command to be executed.                         *
*                         *  *  *  *                       *
* Tries to guess if this is a folder, a URL, etc.  This is *
* not guaranteed to be 100% accurate, or even 6% accurate. *
* This is not extremely efficient either.                  *
***********************************************************/
enum GuessType GuessAnalysis(const char *sCommand)
{
  char token[MAX_LINE_LENGTH], *a, *b;
  DWORD dwAttr;

  if (!GetToken(sCommand, token, NULL, TRUE))
    return eGuessDunno; /* oh, it's probably nothing */ 

  /* well, probably only bangs start with bang */ 
  if (token[0] == '!')
    return eGuessBang;

  a = strchr(token, '/');
  if (a)
  {
    b = strchr(token, ':');
    if (b && (a-b)==1 && a[1] == '/')
      /* does it have "://" in it, before any other / ? */ 
      return eGuessURL;

    /* something, dot, something, slash might be a URL */ 
    b = strchr(token, '.');
    if (b && b!=token && b < a && (a-b)>1)
      return eGuessURLNoProtocol;
  }

  /* hm...  maybe it's a file? */ 
  dwAttr = GetFileAttributes(token);
  if (dwAttr == INVALID_FILE_ATTRIBUTES)
    /* ok, maybe it's got spaces but it isn't quoted? */ 
    dwAttr = GetFileAttributes(sCommand);
  if (dwAttr != INVALID_FILE_ATTRIBUTES)
  {
    if ((dwAttr & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
      return eGuessFolder;
    return eGuessFile;
  }

  return eGuessDunno;
}

/***********************************************************
* void DirectExecCommand()                                 *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND hWnd;                                          *
*      The window to send to ShellExecute                  *
*    - const char *sCommand;                               *
*      The command.                                        *
*                         *  *  *  *                       *
* Directly executes the command, with no extra expansion,  *
* alias or commandchar lookup, or history list stuff.      *
***********************************************************/
int DirectExecCommand(HWND hWnd, const char *sCommand)
{
  int hError;
  enum GuessType Type = eGuessDunno;

  if (cs->AssumeNetAddress || cs->ExplorePaths || *cs->BrowserPath)
  {
    Type = GuessAnalysis(sCommand);
    /* try to explore it first if it looks like a folder */ 
    if (Type == eGuessFolder && cs->ExplorePaths)
    {
      hError = (int)LSExecuteEx(hWnd, "explore", sCommand, NULL, NULL, SW_NORMAL);
      if (hError > 32)
        return hError;
    }
    /* try to open with browser if it looks like a URL */ 
    if (Type == eGuessURL && *cs->BrowserPath)
    {
      hError = (int)LSExecuteEx(hWnd, "open", cs->BrowserPath, sCommand, NULL, SW_NORMAL);
      if (hError > 32)
        return hError;
    }
  }

  /* this should work for executables and bangs */ 
  hError = (int)LSExecute(hWnd, sCommand, SW_NORMAL);
  if (hError > 32)
    return hError;

  /* this is needed if it's a document or URL */ 
  hError = (int)LSExecuteEx(hWnd, "open", sCommand, NULL, NULL, SW_NORMAL);
  if (hError > 32)
    return hError;

  /* this... seems like a bad idea to me, but the feature was already here */ 
  if (cs->AssumeNetAddress && (Type == eGuessURLNoProtocol || Type == eGuessDunno))
  {
    char *sNetAddr = malloc(strlen(sCommand)*sizeof(*sNetAddr) + sizeof("http://"));
    if (sNetAddr)
    {
      strcpy(sNetAddr, "http://");
      strcat(sNetAddr, sCommand);
      hError = (int)LSExecuteEx(hWnd, "open", sNetAddr, NULL, NULL, SW_NORMAL);
      free(sNetAddr);
      if (hError > 32)
        return hError;
    }
  }

  return hError;
}

/***********************************************************
* void ComplainExecError()                                 *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - int hError;                                         *
*      The error returned from LSExecute[Ex]               *
*    - const char *sLiteralCommand;                        *
*      The literal command before expansion.               *
*    - const char *sExpandedCommand;                       *
*      The actual command to be executed.                  *
*                         *  *  *  *                       *
* Complains about a failed execute operation unless        *
* configured not to.                                       *
***********************************************************/
void ComplainExecError(int hError, const char *sLiteralCommand, const char *sExpandedCommand)
{
  const char *sError;
  HLOCAL pData = NULL;
  DWORD dwError;

  if (hError > 32) return;
  if (cs->NoWarnOnError) return;

  dwError = GetLastError();
  if (hError == 32 && dwError != ERROR_DLL_NOT_FOUND)
    sError = "Invalid bang command or internal error.\n";
  else
  {
    /* get a string from the system, hopefully in the user's language */ 
    FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_FROM_SYSTEM,
      NULL, dwError, 0, (LPTSTR)&pData, 0, NULL);
    sError = (const char *)pData;
  }

  if (sLiteralCommand && sExpandedCommand)
  {
    /* don't want to confuse people */ 
    if (strcmp(sLiteralCommand, sExpandedCommand) == 0)
      sExpandedCommand = NULL;
    else
    {
      Complain("%1%nOriginal: %2%nExpanded: %3", sError, sLiteralCommand, sExpandedCommand);
      if (pData) LocalFree(pData);
      return;
    }
  }

  /* if we only have one, just list that one */ 
  if (sLiteralCommand || sExpandedCommand)
  {
    Complain("%1%nCommand: %2", sError, sLiteralCommand ? sLiteralCommand : sExpandedCommand);
    if (pData) LocalFree(pData);
    return;
  }

  Complain("%1", sError);
  if (pData) LocalFree(pData);
}

/***********************************************************
* void ExecCommand()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *szCommand                               *
*      Pointer to the command to execute                   *
*    - BOOL noaddtohist                                    *
*      This specifies whether the command should be added  *
*      to the history.                                     *
*                         *  *  *  *                       *
* This function checks for a valid command and sends it to *
* the proper place - ParseBang, ParseSearch, or ShellEx    *
***********************************************************/

void ExecCommand(const char *sLiteralCommand, BOOL noaddtohist)
{
  HASH Hash;
  char sOrigCase[MAX_LINE_LENGTH], sKeyCase[MAX_LINE_LENGTH];
	char sExpandedCommand[MAX_LINE_LENGTH];
  const char *pCommand, *pArgs;
  const char *sCommandChar;

  if(!sLiteralCommand) /* NULL command */ 
    return;

  pCommand = Parse_SkipSpace(sLiteralCommand);

  /* strlen()==0 will run all the way to the end of the string */ 
  /* just to see if the terminating NUL is right at the front. */ 

  if (!*pCommand) /* empty command, or just whitespace */ 
    return;

  if(!noaddtohist)
    History_Add(sLiteralCommand);

  /* check for command chars first so we can use $ */ 
  if (IS_COMMAND_CHAR(*pCommand))
  {
    sCommandChar = cs->CommandChars[COMMAND_CHAR_INDEX(*pCommand)];
    if (sCommandChar && *sCommandChar)
    {
      if (Parse_ReplaceArgs(sCommandChar, ++pCommand, sExpandedCommand, MAX_LINE_LENGTH))
        ComplainExecError(DirectExecCommand(hWnd, sExpandedCommand), sLiteralCommand, sExpandedCommand);
      return;
    }
  }

  /* expand variables before checking for aliases, etc.             */ 
  /* this happens again in LSExecute, but hopefully it won't matter */ 
  VarExpansion(sExpandedCommand, sLiteralCommand);

  /* copy off the first token. */ 
  Parse_GetHashedToken(sExpandedCommand, &pArgs, sOrigCase, NULL, &Hash, sKeyCase, NULL);

  /* if there's an alias, apply that */ 
  if (Alias_Execute(Hash, sKeyCase, pArgs))
  {
    /* alias executed in Alias_Execute */ 
    return;
  }

  ComplainExecError(DirectExecCommand(hWnd, sExpandedCommand), sLiteralCommand, sExpandedCommand);
}


/***********************************************************
* void DoContextMenu()                                     *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const POINT *ppt;                                   *
*      A pointer to the POINT where the menu should appear *
*      or NULL to choose automatically.                    *
*                         *  *  *  *                       *
* Updates the popup menu, displays it, and performs the    *
* user-selected action.                                    *
***********************************************************/
void DoContextMenu(const POINT *ppt)
{
  POINT pt = { 0, 0 };
  char sText[MAX_LINE_LENGTH];
  HGLOBAL hClipboardText;
  char *sSel, *sClipboard;
  DWORD nStart, nEnd;
  MENUITEMINFO mi;
  UINT id;

  if (!ppt)
  {
    ppt = &pt;
    ClientToScreen(hText, &pt);
  }

  History_CreateMenu(cs);

  id = TrackPopupMenu(hContextMenu,
    TPM_LEFTALIGN | (cs->ContextMenuAboveBox ? TPM_BOTTOMALIGN : TPM_TOPALIGN)
    | TPM_NONOTIFY | TPM_RETURNCMD | TPM_RIGHTBUTTON,
    ppt->x, ppt->y, 0, hWnd, NULL);

  if ((id & HMI_USER_MASK) == 0)
  {
    switch (id)
    {
    case HMI_USER_CUT:
    case HMI_USER_COPY:
      {
        if(!OpenClipboard(hWnd))
          return;
        GetWindowText(hText, sText, sizeof(sText)/sizeof(*sText));
        SendMessage(hText, EM_GETSEL, (WPARAM)&nStart, (LPARAM)&nEnd);
        sSel = sText + nStart;
        sText[nEnd] = '\0';
        sClipboard = GlobalAlloc(GMEM_FIXED, (nEnd - nStart + 1)*sizeof(*sText));
        if (sClipboard)
        {
          memcpy(sClipboard, sSel, (nEnd - nStart + 1)*sizeof(*sText));
          hClipboardText = GlobalHandle(sClipboard);
          SetClipboardData(CF_TEXT, hClipboardText);
          GlobalFree(hClipboardText);
        }
        CloseClipboard();
        if(id == HMI_USER_CUT)
          SendMessage(hText, EM_REPLACESEL, TRUE, (LPARAM)"");
      }
      break;

    case HMI_USER_PASTE:
      {
        if(!OpenClipboard(hWnd))
          return;
        hClipboardText = (void *)GetClipboardData(CF_TEXT);
        if (hClipboardText)
        {
          sClipboard = GlobalLock(hClipboardText);
          if (sClipboard)
          {
            SendMessage(hText, EM_REPLACESEL, TRUE, (LPARAM)sClipboard);
            GlobalUnlock(hClipboardText);
          }
        }
        CloseClipboard();
      }
      break;
    }
  }
  else
  {
    mi.cbSize = sizeof(mi);
    mi.fMask = MIIM_TYPE;
    mi.dwTypeData = sText;
    mi.cch = sizeof(sText)/sizeof(*sText);

    switch (id & HMI_USER_MASK)
    {
    case HMI_USER_ALIAS:
      GetMenuItemInfo(hAliasMenu, id, FALSE, &mi);
      break;

    case HMI_USER_HISTORY:
      GetMenuItemInfo(hHistoryMenu, id, FALSE, &mi);
      break;

    case HMI_USER_SEARCH:
      sText[0] = '?';
      mi.dwTypeData = &sText[1];
      mi.cch -= 2;
      GetMenuItemInfo(hSearchMenu, id, FALSE, &mi);
      strcat(sText, " "); /* will have space because of mi.cch -= 2; */ 
      SetFocusWithText(sText);
      return;

    default:
      return;
    }

    if (cs->ContextMenuExecute)
      ExecCommand(sText, FALSE);
    else
      SetFocusWithTextSelected(sText);
  }
}

#endif //LSXCOMMANDCLOCK_EXPORTS


/***********************************************************
* BOOL CALLBACK EditProc()                                 *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND hText                                          *
*      A handle to a window which this function handles    *
*    - UINT msg                                            *
*      The message sent to the textbox                     *
*    - WPARAM wParam                                       *
*      The WPARAM of this message                          *
*    - LPARAM lParam                                       *
*      The LPARAM of this message                          *
*                         *  *  *  *                       *
* Handles the window procedure for the command textbox.    * 
***********************************************************/

BOOL CALLBACK EditProc(HWND hText,UINT msg,WPARAM wParam,LPARAM lParam)
{
  unsigned cchBuffer, cchPrefix;
  const char *p;
  char *sBuffer;
  int Direction = 7054;

	switch(msg) {
#ifndef LSXCOMMANDCLOCK_EXPORTS
	case WM_KEYDOWN:
    switch (wParam)
    {
      /* evil, ne? */ 
    case VK_DOWN:
      Direction = BROWSE_FORWARD;
      if (0)
      {
    case VK_UP:
        Direction = BROWSE_BACK;
        if (0)
        {
    case VK_NEXT: /* pgdn */ 
          Direction = BROWSE_PAGEFWD;
          if (0)
          {
    case VK_PRIOR: /* pgup */ 
            Direction = BROWSE_PAGEBACK;
            if (0)
            {
    case VK_HOME:
              /* ctrl+home, but don't mess with home */ 
              if (!(GetKeyState(VK_CONTROL)&0x8000))
                break;
              Direction = BROWSE_HOME;
              if (0)
              {
    case VK_END:
                /* ctrl+end, but don't mess with end */ 
                if (!(GetKeyState(VK_CONTROL)&0x8000))
                  break;
                Direction = BROWSE_END;
              }
            }
          }
        }
      }
/* 
      UnixHistory on:
        up = backward, down = forward
      UnixHistory off:
        up = forward, down = backward
      (hm, UnixHistory off seems wierd, and I don't even use Unix shells often)
*/
      if (cs->UnixHistory) Direction = -Direction;
      History_Browse(Direction);
      p = History_GetCurrent();
      if (p)
        SetSelectedText(p);
      return 0;

    case VK_ESCAPE:
      History_Browse(BROWSE_HOME);
      ClearText();
      return 0;

    case VK_TAB:
/*
    FileComplete on:
      TAB: file complete forward
      SHIFT-TAB: file complete backward
      CTRL-TAB: microcomplete

    FileComplete off:
      TAB: microcomplete

    Microcomplete backward might be nice, but how do you stop where
    the microcompletion began?  Going back all the way would be weird.
*/
      if (cs->TabFileComplete && (GetKeyState(VK_CONTROL)&0x8000))
      {
        cchPrefix = GetWindowTextLength(hText);
        cchBuffer = cchPrefix + MAX_PATH + 1;
        sBuffer = malloc(cchBuffer * sizeof(*p));
        if (sBuffer)
        {
          GetWindowText(hText, sBuffer, cchBuffer);
          cchBuffer = Files_Complete(sBuffer, cchBuffer, cchPrefix,
                        (GetKeyState(VK_SHIFT)&0x8000) ? -1 : 1);
          if (cchBuffer != cchPrefix)
          {
            SetWindowText(hText, sBuffer);
            SendMessage(hText, EM_SETSEL, cchPrefix, cchBuffer);
          }
          free(sBuffer);
        }
      }
      else
      {
        if(!cs->NoTabMicroComplete)
        {
          DWORD nStart, nEnd;
          SendMessage(hText, EM_GETSEL, (WPARAM)&nStart, (LPARAM)&nEnd);

          if (nStart != nEnd)
          {
            cchBuffer = GetWindowTextLength(hText) + 1;
            sBuffer = malloc(cchBuffer * sizeof(*p));
            if (sBuffer)
            {
              GetWindowText(hText, sBuffer, cchBuffer);
              p = sBuffer + nStart;
              /* tried GetToken to include quotes, but I don't think that's desireable */ 
              while (*p && !isspace((unsigned char)*p)) ++p;
              while (isspace((unsigned char)*p)) ++p;
              nStart = p - sBuffer;
              SendMessage(hText, EM_SETSEL, (WPARAM)nStart, (LPARAM)nEnd);
              free(sBuffer);
            }
          }
        }
      }
      return 0;
    }
		break;

	case WM_CHAR:
    /* If not TAB, we don't want any file autocompletes. */ 
    if(wParam != VK_TAB)
      Files_Close();

    switch (wParam)
    {
    case VK_RETURN:
      cchBuffer = GetWindowTextLength(hText) + 1;
      sBuffer = malloc(cchBuffer * sizeof(*p));
      if (sBuffer)
      {
        GetWindowText(hText, sBuffer, cchBuffer);
        if (!cs->NoClearOnCommand)
          ClearText();
			  if (cs->HideOnCommand)
				  ShowWindow(hWnd, SW_HIDE);
			  ExecCommand(sBuffer, FALSE);
        free(sBuffer);
      }
			return 0;

    case VK_ESCAPE:
      if (cs->HideOnEscape)
        ShowWindow(hWnd, SW_HIDE);
      return 0;

    case VK_TAB:
      return 0;
    }

/*  huh?  why?
      if(cs->CommaDelimiter && (wParam == 0x2E)) { // Period
        GetWindowText(hText, buf, sizeof(buf));
        if(buf[0] == '=')
          wParam = 0x2C; // Comma
      }
*/
/*  I don't completely understand this code, so I will ignore it and
    write code that will work the way I think autocompletion should.

      if(cs->AutoComplete)
      {
        DWORD nStart, nEnd;
        SendMessage(hText, EM_GETSEL, (WPARAM)&nStart, (LPARAM)&nEnd);
        GetWindowText(hText, buf, sizeof(buf));
        buf2 = (char *)malloc(strlen(buf) + 2);
        buf3 = buf + nEnd*sizeof(char);
        strcpy(buf2, "");
        strncpy(buf2, buf, nStart);
        buf2[nStart] = (char)wParam;
        buf2[nStart + 1] = '\0';
        strcat(buf2, buf3);
        HistoryMoveNext(&current, 1);
        nStart = AutoComplete(hText, buf2);
        if(nStart) {
          free(buf2);
          return 0;
        } else HistoryMoveToTail(&current);  // We got to something that is not in the history.
      }
*/
    if (cs->AutoComplete)
    {
      DWORD nStart, nEnd;
      BOOL bTextSet = FALSE;

      cchPrefix = GetWindowTextLength(hText);
      SendMessage(hText, EM_GETSEL, (WPARAM)&nStart, (LPARAM)&nEnd);
      ASSERT(nStart <= nEnd);
      ASSERT(nEnd <= cchPrefix);
      if (nEnd == cchPrefix)
      {
        cchBuffer = cchPrefix + MAX_LINE_LENGTH + 2;
        sBuffer = malloc(cchBuffer * sizeof(*sBuffer));
        if (sBuffer)
        {
          ASSERT(WRITEABLEN(sBuffer, cchBuffer));
          GetWindowText(hText, sBuffer, cchBuffer);
          /* replace the trailing selection with new char */ 
          cchPrefix = nStart;
          ASSERT(cchPrefix < cchBuffer);
          sBuffer[cchPrefix++] = (char)wParam;
          ASSERT(cchPrefix < cchBuffer);
          sBuffer[cchPrefix] = 0;
          /* autocomplete */ 
          cchBuffer = History_Complete(sBuffer, cchBuffer, cchPrefix);
          ASSERT(cchBuffer >= cchPrefix);
          if (cchBuffer > cchPrefix)
          {
            SetWindowText(hText, sBuffer);
            bTextSet = TRUE;
            SendMessage(hText, EM_SETSEL, cchPrefix, cchBuffer);
          }
          free(sBuffer);
          if (bTextSet)
            return 0;
        }
      }
    }
		break;
	case WM_DROPFILES:
		{
		  char szFileName[MAX_PATH];
      Files_Close();
		  DragQueryFile((HDROP)wParam, 0, szFileName, sizeof(szFileName));
      SetFocusWithTextSelected(szFileName);
		  DragFinish((HDROP)wParam);
		  return 0;
		}
    break;
#endif //LSXCOMMANDCLOCK_EXPORTS
  case WM_LBUTTONDOWN:
#ifndef LSXCOMMANDCLOCK_EXPORTS
    if (GetFocus() != hText)
    {
      SetFocusMouse();
      return 0;
    }
#else //LSXCOMMANDCLOCK_EXPORTS
    {
      HWND active = GetForegroundWindow();
      if(curTimer == WINAMP_TIMER)
      {
        KillTimer(hWnd, ID_WINAMPTIMER);
        TimerProc(hWnd, 0, ID_CLOCKTIMER, 0);
        SetTimer(hWnd, ID_CLOCKTIMER, 1000, TimerProc);
        curTimer = CLOCK_TIMER;
      }
      else
        if(cs->ScrollWinAmp)
          SetTimer(hWnd, ID_WINAMPTIMER, 100, WinAmpTimerProc);

      SetForegroundWindow(active);
      return 0;
    }
#endif //LSXCOMMANDCLOCK_EXPORTS
    break;

  case WM_SHOWWINDOW:
    if (lParam == 0)
    { /* called because of ShowWindow, etc */ 
      if (wParam)
      { /* showing */ 
      }
      else
      { /* hiding */ 
#ifndef LSXCOMMANDCLOCK_EXPORTS
        if(cs->ClearOnHide) ClearText();
#endif /* LSXCOMMANDCLOCK_EXPORTS */ 
      }
    }
    break;

  case WM_SETFOCUS:
    TriggerEvent(EventFocus);
#ifndef LSXCOMMANDCLOCK_EXPORTS
    if ((*(cs->Clock) && cs->ClockDisappears)
        || (cs->ScrollWinAmp && cs->WinAmpDisappears))
        /* && FindWindow(cs->MusicPlayer, NULL) ? */ 
      ClearText();
#endif //LSXCOMMANDCLOCK_EXPORTS
    break;

  case WM_KILLFOCUS:
    /* MSDN: While processing this message, do not make any function calls that display or  */ 
    /* activate a window. This causes the thread to yield control and can cause the applic- */ 
    /* ation to stop responding to messages.                                                */ 
    TriggerEvent(EventUnfocus);
    /* http://msdn.microsoft.com/library/en-us/winui/winui/windowsuserinterface/windowing/windows/windowfeatures.asp#visibility */ 
    /* Hidden windows cannot have input focus, but the system will send WM_KILLFOCUS if it  */ 
    /* becomes hidden while it has focus.  When this happens, it will set the focus hWnd to */ 
    /* NULL (no window has keyboard focus), so we should be able to just check for that.    */ 
    if(cs->HideOnUnfocus && (HWND)wParam != NULL)
	    ShowWindow(hWnd, SW_HIDE);
    break;

#ifndef LSXCOMMANDCLOCK_EXPORTS
  case WM_CONTEXTMENU:
    {
      POINT pt = { LOWORD(lParam), HIWORD(lParam) };
      if (pt.x == (WORD)-1 && pt.y == (WORD)-1)
        /* from keyboard, use default position */ 
        DoContextMenu(NULL);
      else
        DoContextMenu(&pt);
      return 0;
    }
#endif //LSXCOMMANDCLOCK_EXPORTS
  case WM_CLOSE:
    return 0;
	}
	return CallWindowProc(wpOld,hText,msg,wParam,lParam);
}

/***********************************************************
* HBRUSH GetBGBrush()                                      *
*                         *  *  *  *                       *
* Returns the background brush for the edit control,       *
* resizing and recreating the pattern brush if necessary.  *
***********************************************************/
HBRUSH GetBGBrush(void)
{
  static SIZE sBGBrush = { 0, 0 };

  if (hBGBitmap)
  {
    /* if it's the same size, used the cached brush */ 
    if ( sBGBrush.cx != cs->width
      || sBGBrush.cy != cs->height)
    {
      /* maybe we're tiling and haven't created the brush yet? */ 
      if (!cs->m_bStretchImage)
      {
        if (!hBGBrush)
          hBGBrush = CreatePatternBrush(hBGBitmap);
        sBGBrush.cx = cs->width;
        sBGBrush.cy = cs->height;
      }
      else
      {
        /* no such luck, need to recreate the brush */ 
        HDC hTextDC = GetDC(hText);
        HDC hScaledDC = CreateCompatibleDC(hTextDC);
        HDC hBitmapDC = CreateCompatibleDC(hTextDC);
        HBITMAP hScaledBmp = CreateCompatibleBitmap(hTextDC, cs->width, cs->height);

        HGDIOBJ hOldBmpB = SelectObject(hBitmapDC, hBGBitmap);
        HGDIOBJ hOldBmpS = SelectObject(hScaledDC, hScaledBmp);

        BOOL bBlitOk;

        ReleaseDC(hText, hTextDC);

		    SetStretchBltMode(hScaledDC, STRETCH_DELETESCANS);
		    bBlitOk = StretchBlt(hScaledDC, 0, 0, cs->width, cs->height, hBitmapDC, 0, 0, cs->m_nWidth, cs->m_nHeight, SRCCOPY );

        SelectObject(hBitmapDC, hOldBmpB);
        DeleteDC(hBitmapDC);
        SelectObject(hScaledDC, hOldBmpS);
        DeleteDC(hScaledDC);

        if (bBlitOk)
        {
          DeleteObject(hBGBrush);
          hBGBrush = CreatePatternBrush(hScaledBmp);

          sBGBrush.cx = cs->width;
          sBGBrush.cy = cs->height;
        }

        /* the brush has its own copy */ 
        DeleteObject(hScaledBmp);
      }
    }
  }
  else
  {
    /* solid color brush, resizes don't matter */ 
    if (!hBGBrush)
      hBGBrush = CreateSolidBrush(cs->BGColor);
  }

  return hBGBrush;
}

/***********************************************************
* BOOL CALLBACK WndProc()                                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND hWnd                                           *
*      A handle to a textbox which this function handles   *
*    - UINT msg                                            *
*      The message sent to the textbox                     *
*    - WPARAM wParam                                       *
*      The WPARAM of this message                          *
*    - LPARAM lParam                                       *
*      The LPARAM of this message                          *
*                         *  *  *  *                       *
* Handles the window procedure for the main window         * 
***********************************************************/

BOOL CALLBACK WndProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg) {
	case WM_CREATE:
		{
		//hFont = CreateFont(cs->TextSize,0,0,0,0,0,0,0,DEFAULT_CHARSET,0,0,0,0,cs->TextFontFace);
		hFont = CreateFont(
			cs->TextSize,				// height of font
			0,							// average character width
			0,							// angle of escapement
			0,							// base-line orientation angle
			(cs->m_bTextBold ? FW_BOLD: FW_NORMAL),	// font weight
			cs->m_bTextItalic,							// italic attribute option
			cs->m_bTextUnderline,					    // underline attribute option
			0,							// strikeout attribute option
			DEFAULT_CHARSET,			// character set identifier
			0,							// output precision
			0,							// clipping precision
			0,							// output quality
			0,							// pitch and family
			cs->TextFontFace			// typeface name
		);
		hText = CreateWindowEx(0, "EDIT", "", WS_CHILD|ES_AUTOHSCROLL|cs->TextAlign,
      cs->LeftBorderSize, cs->TopBorderSize,
      cs->width-(cs->LeftBorderSize+cs->RightBorderSize),
      cs->height-(cs->TopBorderSize+cs->BottomBorderSize),
      hWnd, 0, hInst, 0);

		hBGBitmap = *(cs->Background) ? LoadLSImage(cs->Background, NULL) : NULL;

		hHollowBrush = (HBRUSH)GetStockObject(HOLLOW_BRUSH);
		if( cs->m_bStretchImage )
			GetLSBitmapSize ( hBGBitmap, &cs->m_nWidth, &cs->m_nHeight );
		if(!hText)
    {
      Complain("Error creating edit window.");
		}
		wpOld = (WNDPROC)SetWindowLong(hText,GWL_WNDPROC,(long)EditProc);
		SendMessage(hText, WM_SETFONT, (WPARAM)hFont, FALSE);
		ShowWindow(hText, SW_SHOWNA);
		SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);
		return 0;
		}
	case WM_SETCURSOR:
		if(cs->NoCursorChange) return TRUE;
		else break;
	case WM_NCHITTEST:
		if(cs->notmoveable) return HTNOWHERE;
		else return HTCAPTION;
	case WM_CTLCOLOREDIT:
    {
      const HDC hDC = (HDC)wParam;
      SetTextColor(hDC, cs->TextColor);
      SetBkColor(hDC, cs->BGColor);
      SetBrushOrgEx(hDC, -cs->LeftBorderSize, -cs->TopBorderSize, NULL);
      SetBkMode(hDC, cs->Transparent?TRANSPARENT:OPAQUE);
      return (int)GetBGBrush();
    }
    break;
  case WM_PAINT:
    if(hBGBitmap)
    {
      PAINTSTRUCT ps;
      HDC src = CreateCompatibleDC(NULL);
      HDC hdc = BeginPaint(hWnd, &ps);
      SelectObject(src, hBGBitmap);
	    if( cs->m_bStretchImage )
      {
		    SetStretchBltMode( hdc, STRETCH_DELETESCANS ); 
		    if( !StretchBlt( hdc, 0, 0, cs->width, cs->height, src, 0, 0, cs->m_nWidth, cs->m_nHeight, SRCCOPY ) )
			    break;
	    }
	    else
      {
		    if(!BitBlt(hdc, 0, 0, cs->width, cs->height, src, 0, 0, SRCCOPY))
			    break;
		    //TransparentBltLS((HDC)wParam, 0, 0, cs->width, cs->height, src, 0, 0, RGB(255,0,255));
	    }
      EndPaint(hWnd, &ps);
      DeleteDC(src);
      return 0;
    }
    break;
  case WM_DESTROY:
	    //20021115
		SendMessage(GetLitestepWnd(), LM_UNREGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);
        if (lsboxed)
		{
			cs->WaitForBox = TRUE;
			SetWindowLong(hWnd, GWL_STYLE, (GetWindowLong(hWnd, GWL_STYLE) &~ WS_CHILD)|WS_POPUP);
			SetParent(hWnd, 0);
			SetWindowPos(hWnd,HWND_TOP,cs->x,cs->y,0,0,SWP_NOSIZE|SWP_HIDEWINDOW);
			lsboxed = FALSE;
			return 0;
		}
    break;
#ifndef LSXCOMMANDCLOCK_EXPORTS
	case WM_DROPFILES:
		{
		char szFileName[MAX_PATH];
		DragQueryFile((HDROP)wParam, 0, szFileName, sizeof(szFileName));
    SetFocusWithTextSelected(szFileName);
		DragFinish((HDROP)wParam);
		return 0;
		}
#endif //LSXCOMMANDCLOCK_EXPORTS
	//20021115
	case LM_GETREVID:
		{
			char* buf = (char*)(lParam);
	
			switch (wParam)
			{
				case 0:
				case 1:
					sprintf(buf, "LsxCommand %s ", rcsRevision);
					break;
				default:
					strcpy(buf, "");
			}
			return strlen(buf);
		}

  case CM_TRIGGEREVENT:
    {
      const char *sCommand = NULL;
      switch (wParam)
      {
      case EventFocus:    sCommand = cs->OnFocusCommand; break;
      case EventUnfocus:  sCommand = cs->OnUnfocusCommand; break;
      }
      if (sCommand != NULL)
        LSExecute(hWnd, sCommand, SW_NORMAL);
    }
    return TRUE;
	}
	return DefWindowProc(hWnd,msg,wParam,lParam);
}


/***********************************************************
* VOID CALLBACK TimerProc()                                *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND hWnd                                           *
*      The HWND of the window that owns this timer.        *
*    - UINT uMsg                                           *
*      The WM_TIMER message.                               *
*    - UINT idEvent                                        *
*      ID of the Timer event.                              *
*    - DWORD dwTime                                        *
*      Current system time.                                *
*                         *  *  *  *                       *
* If the user chooses to have a clock in LSXCommand, this  *
* Callback processes the request to update the time & date *
***********************************************************/

VOID CALLBACK TimerProc(HWND hWnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
#ifndef LSXCOMMANDCLOCK_EXPORTS
  if(GetFocus() != hText && IsWindowVisible(hWnd) && !cs->WaitForBox) {
#else
  if(IsWindowVisible(hWnd)) {
#endif
    time_t tVal;
    struct tm *stVal;
    char tstring[512];

    time(&tVal);
    stVal = localtime(&tVal);
    strftime(tstring, sizeof(tstring), cs->Clock, stVal);

    SetWindowText(hText, tstring);
#ifndef LSXCOMMANDCLOCK_EXPORTS
    Files_Close();
#endif
  }
}


/***********************************************************
* VOID CALLBACK WinAmpTimerProc()                          *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND hWnd                                           *
*      The HWND of the window that owns this timer.        *
*    - UINT uMsg                                           *
*      The WM_TIMER message.                               *
*    - UINT idEvent                                        *
*      ID of the Timer event.                              *
*    - DWORD dwTime                                        *
*      Current system time.                                *
*                         *  *  *  *                       *
* If the user chooses to display the current WinAmp title, *
* this function updates the lsxcommand display.            *
***********************************************************/

VOID CALLBACK WinAmpTimerProc(HWND hWnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
  char title[MAX_PATH], curr[MAX_PATH];

#ifndef LSXCOMMANDCLOCK_EXPORTS
  if (GetFocus() != hText && IsWindowVisible(hWnd))
#else
  if (IsWindowVisible(hWnd))
#endif
  {
    HWND wahWnd = FindWindow(cs->MusicPlayer, NULL);
    if(wahWnd)
    {
      if(curTimer == CLOCK_TIMER)
      {
        KillTimer(hWnd, ID_CLOCKTIMER);
        curTimer = WINAMP_TIMER;
      }
      GetWindowText(wahWnd, title, sizeof(title));
      GetWindowText(hText, curr, sizeof(curr));
      if (strcmp(title, curr))
        SetWindowText(hText, title);
#ifndef LSXCOMMANDCLOCK_EXPORTS
      Files_Close();
#endif
    }
    else if(*(cs->Clock) && curTimer != CLOCK_TIMER)
    {
      SetTimer(hWnd, ID_CLOCKTIMER, 1000, TimerProc);
      curTimer = CLOCK_TIMER;
    }
  }
}


/***********************************************************
* void BangFocusCommand()                                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Handles the !Bang Command to focus the command box       *
***********************************************************/

void BangFocusCommand(HWND caller, const char *args)
{
  SetFocusKeyboard();
}


/***********************************************************
* void BangToggleCommand()                                 *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Handles the !Bang Command to toggle the visibility of    *
* the command box.                                         *
***********************************************************/

void BangToggleCommand(HWND caller, const char *args)
{
	if(IsWindowVisible(hWnd))
		ShowWindow(hWnd, SW_HIDE);
	else 
    SetFocusKeyboard();
	return;
}


/***********************************************************
* void BangRescanEngineList()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Handles the !Bang Command to rescan the list of search   *
* engines.                                                 *
***********************************************************/

#ifndef LSXCOMMANDCLOCK_EXPORTS
void BangRescanEngineList(HWND caller, const char *args)
{
  Search_Unload(cs);
  Alias_Unload(cs);
  Search_Load(cs);
  Alias_Load(cs);

  History_CreateMenu(cs);

  ContextMenuInit();
}


/***********************************************************
* void BangClearHistory()                                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Handles the !Bang Command to rescan the list of search   *
* engines.                                                 *
***********************************************************/

void BangClearHistory(HWND caller, const char *args)
{
  History_Clear(cs);
}


/***********************************************************
* void BangShowContextMenu()                               *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Makes LSXCommand's Context Menu visible wherever the     *
* cursor is on the screen, regardless of LSXCommand's      *
* visibility.                                              *
***********************************************************/

void BangShowContextMenu(HWND caller, const char *args)
{
  POINT pt;
  GetCursorPos(&pt);
  DoContextMenu(&pt);
}


/***********************************************************
* void BangExecCommand()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Executes a command.  This enables any part of Litestep   *
* to use the power of LSXCommand's command engine.         *
* Changed from BangCommand(), since that's now a datatype  *
* inside lsapi.h in newer builds.                          *
***********************************************************/

void BangExecCommand(HWND caller, const char *args)
{
  ExecCommand(args, !(cs->AddExternalsToHistory));
}

/***********************************************************
* void BangExecCalculator()                                *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Evaluates an expression and pastes the result into the   *
* command box.  This is used to provide the =... functions *
* starting with lsxcommand 1.8.6                           *
***********************************************************/

const char *sBangExecCalculator = "!CommandCalculate";
void BangExecCalculator(HWND caller, const char *args)
{
  char *newcmd;
  if (!args || !*args) return;

  /* ParseCalculatorCommand needs to modify its input */ 
  newcmd = malloc(strlen(args) + 2);
  newcmd[0] = newcmd[1] = 0;
  if (*args != '=') newcmd[0] = '=';
  strcat(newcmd, args);
  ParseCalculatorCommand(newcmd);
  free(newcmd);
}

/***********************************************************
* void BangExecSearch()                                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Perform a search, used to provide the ?... functionality *
* starting with lsxcommand 1.8.6                           *
***********************************************************/
const char *sBangExecSearch = "!CommandSearch";
void BangExecSearch(HWND caller, const char *args)
{
  char OrigCase[MAX_LINE_LENGTH], KeyCase[MAX_LINE_LENGTH];
  HASH hash;

  *KeyCase = 0;
  /* "? <args>" or "?<engine> <args>" or "<args>" */ 
  if (*args == '?')
  {
    ++args;
    if (!isspace((unsigned char)*args))
      Parse_GetHashedToken(args, &args, OrigCase, NULL, &hash, KeyCase, NULL);
  }

  Search_Execute(hWnd, hash, KeyCase, args);
}

#endif //LSXCOMMANDCLOCK_EXPORTS

/***********************************************************
* void BangShowCommand()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Just shows the command box if it isn't visible.          *
***********************************************************/

void BangShowCommand(HWND caller, const char *args)
{
  if(!IsWindowVisible(hWnd))
		ShowWindow(hWnd, SW_SHOWNA);
  /* What's the harm of showing it if it is visible? [rc] */ 
}


/***********************************************************
* void BangHideCommand()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Just hides the command box if it is visible.             *
***********************************************************/

void BangHideCommand(HWND caller, const char *args)
{
  if(IsWindowVisible(hWnd))
		ShowWindow(hWnd, SW_HIDE);
}


#ifndef LSXCOMMANDCLOCK_EXPORTS
/***********************************************************
* void BangBrowseFile()                                    *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* Opens up an "Open File" dialog box so the user doesn't   *
* have to type...                                          *
***********************************************************/

void BangBrowseFile(HWND caller, const char *args)
{
  char file[MAX_PATH], file_total[MAX_PATH], filetitle[MAX_PATH], title[] = "LSXCommand Open File", *filter = NULL, *p = NULL, *p2 = NULL;
  OPENFILENAME ofn;

  if(args && *args) {
    filter = (char *)malloc(strlen(args) + 2);
    strcpy(filter, args);
    filter[strlen(filter) + 1] = '\0';
    p2 = filter;
    p = strchr(p2, '|');
    while(p) {
      *p = '\0';
      p2 += strlen(p2) + 1;
      p = strchr(p2, '|');
    }
  } else {
    filter = (char *)malloc(strlen("All Files (*.*)\0*.*") + 1);
    strcpy(filter, "All Files (*.*)\0*.*");
  }

  strcpy(file, "*.*\0");
  ofn.lStructSize = sizeof(OPENFILENAME);
  ofn.hwndOwner = hWnd;
  ofn.lpstrFilter = filter;
  ofn.lpstrCustomFilter = (LPSTR) NULL;
  ofn.nMaxCustFilter = 0L;
  ofn.nFilterIndex = 0;
  ofn.lpstrFile = file;
  ofn.nMaxFile = sizeof(file);
  ofn.lpstrFileTitle = filetitle;
  ofn.nMaxFileTitle = sizeof(filetitle);
  ofn.lpstrInitialDir = NULL;
  ofn.nFileOffset = 0;
  ofn.nFileExtension = 0;
  ofn.lpstrDefExt = "";
  ofn.lpstrTitle = title;
  ofn.Flags = OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST | OFN_ALLOWMULTISELECT | OFN_HIDEREADONLY | OFN_EXPLORER; /*OFN_EXPLORER; */

  if(GetOpenFileName(&ofn)) {
    p = ofn.lpstrFile;
    strcpy(file_total, p);
    if(file_total[strlen(file_total) - 1] == '\\') {
      p2 = file_total + strlen(file_total);

      p += strlen(p) + 1;
      while(*p) {
        strcpy(p2, p);
        ExecCommand(file_total, TRUE);
        p += strlen(p) + 1;
      }
    } else
      ExecCommand(file_total, TRUE);
  }

  if(filter)
    free(filter);
}


/***********************************************************
* void BangBrowseFolder()                                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Should be the arguments to the function...          *
*                         *  *  *  *                       *
* This lets the user select a folder to browse.            *
***********************************************************/

void BangBrowseFolder(HWND caller, const char *args)
{
  char title[] = "LSXCommand Browse", path[MAX_PATH];
  BROWSEINFO bi = {hWnd, NULL, path, "LSXCommand Browse", BIF_RETURNONLYFSDIRS, NULL, 0, 0};
  LPITEMIDLIST pidl;
  IMalloc *pm;

  bi.hwndOwner = hWnd;
  bi.pidlRoot = NULL;
  bi.pszDisplayName = path;
  bi.lpszTitle = title;
  bi.ulFlags = BIF_RETURNONLYFSDIRS;
  bi.lpfn = NULL;
  bi.lParam = 0;
  bi.iImage = 0;

  pidl = SHBrowseForFolder(&bi);

  if(pidl && SHGetPathFromIDList(pidl, path)) {
    ExecCommand(path, TRUE);
    SHGetMalloc(&pm);
    if (pm) {
      pm->lpVtbl->Free(pm,pidl);
      pm->lpVtbl->Release(pm);
    }
  }
}
#endif //LSXCOMMANDCLOCK_EXPORTS


/***********************************************************
* const char *ParseRelCoord()                              *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - int *pValue                                         *
*      A pointer to the current value, and the resulting   *
*      parsed value.  (Use a temp var if you don't want it *
*      overwritten.)                                       *
*    - const char *args                                    *
*      The argument string to parse from.  Preceeding      *
*      whitespace will be trimmed by this function.        *
*    - int vMin                                            *
*      The minimum value, for use with %'s.                *
*    - int vMax                                            *
*      The maximum value, for use with %'s.  If both vMin  *
*      and vMax are 0, %'s are not allowed, nor is C       *
*      Limits are NOT strictly enforced.                   *
*                         *  *  *  *                       *
*    Returns:                                              *
*                                                          *
*    - const char *                                        *
*      A pointer to the first char in args after the token *
*      that this function uses, or NULL if parsing fails.  *
*                         *  *  *  *                       *
* Converts a token in a string to an integer value, allow- *
* ing centered, percent-based, or relative values.         *
* [blame rabidcow, added 2004-02-08]                       *
***********************************************************/

const char *ParseRelCoord(int *pValue, const char *args, int vMin, int vMax)
{
  int sign = 1;
  int vParsed = 0;
  int vAux = 0;

  if (!args || !*args)
    return NULL;

  if (!pValue) /* should never happen */ 
    return NULL;

  /* strip whitespace */ 
  while (*args==' ' || *args=='\t') ++args;
  /* get sign (+- is valid, so you can do 0%+$blah$) */ 
  if (*args=='+') ++args;
  if (*args=='-') { sign = -1; ++args; }
  /* convert to integer */ 
  while (*args>='0' && *args<='9')
    vParsed = vParsed*10 + (*args++)-'0';
  /* interpret trailing char */ 
  switch (*args++)
  {
  case 'c': case 'C':
    if (vMin==0 && vMax==0) return NULL;
    vParsed = (vMin+vMax)/2 + sign*vParsed;
    break;
  case ' ': case '\t': case 0:
    if (vMin==0 && vMax==0)
      vParsed = sign*vParsed;
    else
      vParsed = (sign==-1) ? (vMax - vParsed) : (vMin + vParsed);
    break;
  case 'r': case 'R':
    vParsed = *pValue + sign*vParsed;
    break;
  case '%':
    if (vMin==0 && vMax==0) return NULL;
    vParsed = MulDiv(vParsed, vMax-vMin, 100);
    vParsed = (sign==-1) ? (vMax - vParsed) : (vMin + vParsed);
    /* allow x%+y */ 
    if (*args=='+' || *args=='-')
    {
      args = ParseRelCoord(&vAux, args, 0, 0);
      if (!args) return NULL;
      vParsed += vAux;
    }
    break;
  default:
    return NULL;
  }

  *pValue = vParsed;
  return args;
}

/***********************************************************
* void ComplainBangArgs()                                  *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *sBang                                   *
*      Name of the bang command filing the complaint.      *
*    - const char *sArgs                                   *
*      Original arguments that it's complaining about.     *
*                         *  *  *  *                       *
*  Complains at the user.                                  *
***********************************************************/

void ComplainBangArgs(const char *sBang, const char *sArgs)
{
  Complain("Invalid arguments to %1:%n%2", sBang, sArgs);
}

/***********************************************************
* void BangMove()										                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Space delimited pair of numbers telling where to    *
*      place the main window.                              *
*                         *  *  *  *                       *
* Moves the main LSXCommand according to the input         *
***********************************************************/

const char *sBangMove = "!CommandMove";
void BangMove(HWND caller, const char *args)
{
  const char *origargs = args;
  int oldx = cs->x, oldy = cs->y;

  if (args && *args)
  {
    if (lstrcmpi(args, "home") == 0)
    {
      cs->x = cs->origx;
      cs->y = cs->origy;
    }
    else
    {
      /* limited to primary monitor -- general solution is more work */ 
      /* this is also wrong when docked */ 
      args = ParseRelCoord(&cs->x, args, 0, GetSystemMetrics(SM_CXSCREEN));
      args = ParseRelCoord(&cs->y, args, 0, GetSystemMetrics(SM_CYSCREEN));
      if (!args)
      {
        ComplainBangArgs(sBangMove, origargs);
        return;
      }
    }
  }
  else
  {
    POINT cursorPos;

    GetCursorPos(&cursorPos);

    cs->x = cursorPos.x;
    cs->y = cursorPos.y;
  }

  if(oldx != cs->x || oldy != cs->y) {
    //SetWindowPos(hWnd, cs->NoAlwaysOnTop ? HWND_NOTOPMOST : HWND_TOPMOST, cs->x, cs->y, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);
    /* and this may be wrong when onbottom with multiple monitors */ 
    MoveWindow(hWnd, cs->x, cs->y, cs->width, cs->height, TRUE);
  }

  if(!origargs || !(*origargs))
    BangFocusCommand(caller, NULL);
}


/***********************************************************
* void BangSize()										                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Space delimited pair of numbers telling what size   *
*      make the main window.                               *
*                         *  *  *  *                       *
* Resizes the main LSXCommand window according to input    *
***********************************************************/

const char *sBangSize = "!CommandSize";
void BangSize(HWND caller, const char *args)
{
  const char *origargs = args;
  int oldwidth = cs->width, oldheight = cs->height;

  if (!args || !*args)
    return;

  /* you have a better suggestion? */ 
  if (lstrcmpi(args, "home") == 0)
  {
    cs->width = cs->origwidth;
    cs->height = cs->origheight;
  }
  else
  {
    /* limited to primary monitor -- general solution is more work */ 
    /* this is also wrong when docked */ 
    /* DO NOT change the 0 if you fix for multimon, this is a size rather than a position */ 
    args = ParseRelCoord(&cs->width, args, 0, GetSystemMetrics(SM_CXSCREEN));
    args = ParseRelCoord(&cs->height, args, 0, GetSystemMetrics(SM_CYSCREEN));
    if (!args)
    {
      ComplainBangArgs(sBangSize, origargs);
      return;
    }
  }

  if(oldwidth != cs->width || oldheight != cs->height)
  {
    MoveWindow(hText,
      cs->LeftBorderSize, cs->TopBorderSize,
      cs->width-(cs->LeftBorderSize+cs->RightBorderSize),
      cs->height-(cs->TopBorderSize+cs->BottomBorderSize),
      TRUE);
    MoveWindow(hWnd, cs->x, cs->y, cs->width, cs->height, TRUE);
  }
}



/***********************************************************
* void BangToggleTimer()                                   *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Specifies which time to switch to                   *
*                         *  *  *  *                       *
* Toggles which timer is currently active.                 *
***********************************************************/

void BangToggleTimer(HWND caller, const char *args) {

  HWND active = GetForegroundWindow();

  if(curTimer == WINAMP_TIMER) {
    KillTimer(hWnd, ID_WINAMPTIMER);
    TimerProc(hWnd, 0, ID_CLOCKTIMER, 0);
    SetTimer(hWnd, ID_CLOCKTIMER, 1000, TimerProc);
    curTimer = CLOCK_TIMER;
  } else if(cs->ScrollWinAmp) {
      SetTimer(hWnd, ID_WINAMPTIMER, 100, WinAmpTimerProc);
  }

  SetForegroundWindow(active);
}

#ifndef LSXCOMMANDCLOCK_EXPORTS
/***********************************************************
* void BangSetText()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Specifies the text to place in LSXCommand.          *
*                         *  *  *  *                       *
* Replaces the text inside LSXCommand and brings focus to  *
* the end of the line.                                     *
***********************************************************/

void BangSetText(HWND caller, const char *args)
{
  if(args)
    SetFocusWithText(args);
}

/***********************************************************
* void BangShowText()                                      *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Specifies the text to place in LSXCommand.          *
*                         *  *  *  *                       *
* Replaces the text inside LSXCommand without changing     *
* input focus or selecting it.                             *
***********************************************************/

const char *sBangShowText = "!CommandShowText";
void BangShowText(HWND caller, const char *args)
{
  if (args && *args)
    SetWindowText(hText, args);
}

#endif LSXCOMMANDCLOCK_EXPORTS

/***********************************************************
* void BangBoxHook()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND caller                                         *
*      The HWND of the window that called this !Bang       *
*      Command.                                            *
*    - const char *args                                    *
*      Options passed by LsBox to LsxCommand               *
*                         *  *  *  *                       *
***********************************************************/
void BangBoxHook(HWND caller, const char *args)
{
  char *handle = strrchr(args,' ');

  if (handle) 
  {
	  HWND boxwnd = (HWND)atoi(handle+1);
	  if (boxwnd) 
	  {
		  lsboxed = TRUE;

		  if (boxwnd != GetParent(hWnd))
		  {
			  SetWindowLong(hWnd, GWL_STYLE, (GetWindowLong(hWnd, GWL_STYLE) &~ WS_POPUP)|WS_CHILD);
			  SetParent(hWnd, boxwnd);
			  // THis would cause problems in Win2k
			  // SetWindowLong(hWnd, GWL_STYLE, (GetWindowLong(hWnd, GWL_STYLE) &~ WS_CHILD)|WS_POPUP);
			  if (cs->WaitForBox)
				  cs->WaitForBox = FALSE;
			  //20021114
			  if(cs->HiddenOnStart)
				  ShowWindow(hWnd, SW_HIDE);
			  else
				  ShowWindow(hWnd, SW_SHOWNA);
		  }
	  }
  }
	return;
}


/***********************************************************
* void Complain()                                          *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - const char *sFormat                                 *
*      The format string for the message, in FormatMessage *
*      format.  **NOT** printf format.                     *
*                         *  *  *  *                       *
* Complains to the user about some environmental (ie,      *
* config or system) error.                                 *
* Not to be used for internal debugging.                   *
*                                                          *
* Use %n for required line breaks in the format string,    *
* otherwise they will be inserted automatically to fit the *
* message in 60 columns.                                   *
***********************************************************/
void Complain(const char *sFormat, ...)
{
  char *pMessage;
  va_list args;
  va_start(args, sFormat);

  if (FormatMessage(FORMAT_MESSAGE_FROM_STRING|FORMAT_MESSAGE_ALLOCATE_BUFFER|60,
    sFormat, 0, 0, (LPSTR)&pMessage, 0, &args))
  {
    TRACE1(pMessage);
    MessageBox(hWnd, pMessage, "lsxCommand", MB_OK|MB_TOPMOST|MB_ICONERROR);
    LocalFree((HLOCAL)pMessage);
  }
  va_end(args);
}


/***********************************************************
* int initModuleEx()                                       *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND parent                                         *
*      The HWND of the parent window to this module        *
*    - HINSTANCE dll                                       *
*      The HINSTANCE of this DLL.                          *
*    - LPCSTR szPath                                       *
*      A pointer to the Litestep path                      *
*                         *  *  *  *                       *
* Handles all initialization of LSXCommand.  Returns 0     *
***********************************************************/

int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath)
{
	WNDCLASS wc;
	HBRUSH hBr;

  ModuleLogInit("command", 7, "lsxcommand", 10, rcsRevision, sizeof(rcsRevision)-1);
  TRACE1("initModuleEx");

  setlocale(LC_ALL, "");

	hInst = dll;

  cs = ReadSettings(szPath);
  if( !cs )
  {
    return 1;
  }
  ASSERT(READABLE(cs));

	hBr = CreateSolidBrush(cs->BorderColor);
#ifndef LSXCOMMANDCLOCK_EXPORTS
  Search_Load(cs);
  Alias_Load(cs);
  History_Load(cs);
  ContextMenuInit();
  
  if(cs->ClearHistoryOnStartup)
  {
    History_Clear(cs);
  }
#endif //LSXCOMMANDCLOCK_EXPORTS

  memset(&wc,0,sizeof(wc));
	wc.hInstance = hInst;
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.lpszClassName = szApp;
	wc.hbrBackground = hBr;
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	RegisterClass(&wc);

  hWnd = CreateWindowEx(
    ((cs->NoAlwaysOnTop)?0:WS_EX_TOPMOST)|WS_EX_TOOLWINDOW,
    szApp, "", WS_POPUP|((cs->BevelBorder)?WS_SIZEBOX:0),
    cs->x, cs->y, cs->width, cs->height, GetDesktopWindow(), 0, hInst, 0);

  //DragAcceptFiles(hWnd, TRUE);
	SetWindowLong(hWnd,GWL_USERDATA,magicDWord);

#ifndef LSXCOMMANDCLOCK_EXPORTS
  AddBangCommand("!TOGGLECOMMAND",BangToggleCommand);
  AddBangCommand("!FOCUSCOMMAND",BangFocusCommand);
  AddBangCommand("!COMMANDRESCANENGINES", BangRescanEngineList);
  AddBangCommand("!COMMANDCLEARHISTORY", BangClearHistory);
  AddBangCommand("!COMMANDSHOWCONTEXTMENU", BangShowContextMenu);
  AddBangCommand("!COMMAND", BangExecCommand);
  AddBangCommand(sBangExecCalculator, BangExecCalculator);
  AddBangCommand(sBangExecSearch, BangExecSearch);
  AddBangCommand("!COMMANDSHOW", BangShowCommand);
  AddBangCommand("!COMMANDHIDE", BangHideCommand);
  AddBangCommand("!COMMANDBROWSEFILE", BangBrowseFile);
  AddBangCommand("!COMMANDBROWSEFOLDER", BangBrowseFolder);
  AddBangCommand(sBangMove, BangMove);
  AddBangCommand(sBangSize, BangSize);
  AddBangCommand("!COMMANDTOGGLETIMER", BangToggleTimer);
  AddBangCommand("!COMMANDSETTEXT", BangSetText);
  AddBangCommand(sBangShowText, BangShowText);
  AddBangCommand("!COMMANDBOXHOOK", BangBoxHook);
#else //LSXCOMMANDCLOCK_EXPORTS
  AddBangCommand("!TOGGLECOMMANDCLOCK",BangToggleCommand);
  AddBangCommand("!FOCUSCOMMANDCLOCK",BangFocusCommand);
  AddBangCommand("!COMMANDCLOCKSHOW", BangShowCommand);
  AddBangCommand("!COMMANDCLOCKHIDE", BangHideCommand);
  AddBangCommand("!COMMANDCLOCKMOVE", BangMove);
  AddBangCommand("!COMMANDCLOCKTOGGLETIMER", BangToggleTimer);
  AddBangCommand("!COMMANDCLOCKBOXHOOK", BangBoxHook);
#endif

  if(!cs->HiddenOnStart)
  {
	  if(cs->WaitForBox)
		  ShowWindow(hWnd,SW_HIDE);
	  else
		  ShowWindow(hWnd,SW_SHOWNA);
  }
  else
  {
    ShowWindow(hWnd, SW_HIDE);
  }

  if(*(cs->Clock))
  {
    SetTimer(hWnd, ID_CLOCKTIMER, 1000, TimerProc);
    curTimer = CLOCK_TIMER;
  }

  if(cs->ScrollWinAmp)
  {
    SetTimer(hWnd, ID_WINAMPTIMER, 100, WinAmpTimerProc);
  }

  //Initialize input to nothing
  SetWindowText(hText, "");

	return 0;
}


/***********************************************************
* int initModule()                                         *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HWND parent                                         *
*      The HWND of the parent window to this module        *
*    - HINSTANCE dll                                       *
*      The HINSTANCE of this DLL.                          *
*    - WharfDataType *wd                                   *
*      A pointer to the structure containing data about LS *
*                         *  *  *  *                       *
* Handles all initialization of LSXCommand.  Returns 0     *
***********************************************************/

int initModule(HWND parent, HINSTANCE dll, wharfDataType* wd)
{
  return initModuleEx(parent, dll, wd->lsPath);
}


/***********************************************************
* int quitModule()                                         *
*                         *  *  *  *                       *
*    Arguments:                                            *
*                                                          *
*    - HINSTANCE dll                                       *
*      The HINSTANCE of this DLL.                          *
*                         *  *  *  *                       *
* Handles all of the unloading of this DLL.                *
***********************************************************/

int quitModule(HINSTANCE dll)
{
#ifndef LSXCOMMANDCLOCK_EXPORTS
  RemoveBangCommand("!TOGGLECOMMAND");
  RemoveBangCommand("!FOCUSCOMMAND");
  RemoveBangCommand("!COMMANDRESCANENGINES");
  RemoveBangCommand("!COMMANDCLEARHISTORY");
  RemoveBangCommand("!COMMANDSHOWCONTEXTMENU");
  RemoveBangCommand("!COMMAND");
  RemoveBangCommand(sBangExecCalculator);
  RemoveBangCommand(sBangExecSearch);
  RemoveBangCommand("!COMMANDSHOW");
  RemoveBangCommand("!COMMANDHIDE");
  RemoveBangCommand("!COMMANDBROWSEFILE");
  RemoveBangCommand("!COMMANDBROWSEFOLDER");
  RemoveBangCommand(sBangMove);
  RemoveBangCommand(sBangSize);
  RemoveBangCommand("!COMMANDTOGGLETIMER");
  RemoveBangCommand("!COMMANDSETTEXT");
  RemoveBangCommand(sBangShowText);
  RemoveBangCommand("!COMMANDBOXHOOK");
#else //LSXCOMMANDCLOCK_EXPORTS
  RemoveBangCommand("!TOGGLECOMMANDCLOCK");
  RemoveBangCommand("!FOCUSCOMMANDCLOCK");
  RemoveBangCommand("!COMMANDCLOCKSHOW");
  RemoveBangCommand("!COMMANDCLOCKHIDE");
  RemoveBangCommand("!COMMANDCLOCKMOVE");
  RemoveBangCommand("!COMMANDCLOCKTOGGLETIMER");
  RemoveBangCommand("!COMMANDCLOCKBOXHOOK");
#endif

#ifndef LSXCOMMANDCLOCK_EXPORTS
	History_Unload(cs);
  Search_Unload(cs);
  Alias_Unload(cs);
  DestroyMenu(hContextMenu);
#endif //LSXCOMMANDCLOCK_EXPORTS
	DestroyWindow(hWnd);
	UnregisterClass(szApp,hInst);
	DeleteObject(hFont);
  DeleteObject(hBGBrush);
  DeleteObject(hHollowBrush);
  DeleteObject(hBGBitmap);
	//20021114
	if (cs)
	{
    FreeCommandChars(cs);
		if (cs->OnFocusCommand != NULL) {
			//delete [] cs->OnFocusCommand;
			free(cs->OnFocusCommand);
			cs->OnFocusCommand = NULL;
		}
		if (cs->OnUnfocusCommand != NULL) {
			//delete [] cs->OnUnfocusCommand;
			free(cs->OnUnfocusCommand);
			cs->OnUnfocusCommand = NULL;
		}
		free(cs);

	}

  TRACE1("quitModule");
  ModuleLogQuit();

	return 0;
}
