/******************************************************************************
*                                                                             *
* This is a part of the LsxCommand LiteStep plugin Source code.               *
*                                                                             *
* Copyright (C) 1999-2000 Visigoth (Shaheen Gandhi in real life)              *
* Based on limpid's lscommand.                                                *
* Updated by blkhawk, ilmcuts, jesus_mjjg, rabidcow.                          *
* Look at the documentation for more informations.                            *
*                                                                             *
* This program is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by        *
* the Free Software Foundation; either version 2 of the License, or           *
* (at your option) any later version.                                         *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA   *
*                                                                             *
******************************************************************************/

#ifndef LSCOMMAND_H
#define LSCOMMAND_H

#include "resource.h"
#include "aggressiveoptimize.h"

// remove this
#undef _MAX_PATH

/* Timer Enumeration */
enum Timers {
  NO_TIMER = 0,
  CLOCK_TIMER = ID_CLOCKTIMER,
  WINAMP_TIMER = ID_WINAMPTIMER
};

#ifndef LSXCOMMANDCLOCK_EXPORTS

/* Used for Context Menu entries */
#define HMI_USER_CUT      0x0101
#define HMI_USER_COPY     0x0102
#define HMI_USER_PASTE    0x0103
#define HMI_USER_MASK     0x6000
#define HMI_USER_ALIAS    0x2000
#define HMI_USER_HISTORY  0x4000
#define HMI_USER_SEARCH   0x6000

#endif

/* Data Structures */
struct CommandSettings {
	//image sizes
	int m_nWidth;
	int m_nHeight;
	BOOL m_bStretchImage;
	//font
	BOOL m_bTextBold;
	BOOL m_bTextUnderline;
	BOOL m_bTextItalic; 
	//go on with what was before (what does that mean?)
	char TextFontFace[256];
  char SearchEngineList[MAX_PATH];
  char ContextMenuOrder[4];
  char Background[MAX_PATH];
  char Clock[256];
  UINT TextAlign; /* hm, was this here and unused... */ 
  char BrowserPath[MAX_PATH];
  char MusicPlayer[128];

	int TextSize;
	int x, origx;
	int y, origy;
	int width, origwidth;
	int height, origheight;
	int BorderSize;
	int TopBorderSize;
	int BottomBorderSize;
	int LeftBorderSize;
	int RightBorderSize;
  int Transparent;
#ifndef LSXCOMMANDCLOCK_EXPORTS
  int MaxHistoryEntries;
  int MaxHistoryMenuEntries;
  int ContextMenuStandardItems;
#endif // LSXCOMMANDCLOCK_EXPORTS
	BOOL NoCursorChange;
	BOOL BevelBorder;
	BOOL NoAlwaysOnTop;
	BOOL notmoveable;
	BOOL HiddenOnStart;
	BOOL WaitForBox;
  BOOL SelectAllOnFocus;
  BOOL SelectAllOnMouseFocus;
  BOOL ScrollWinAmp;
  BOOL HideOnUnfocus;
  BOOL AddExternalsToHistory;
#ifndef LSXCOMMANDCLOCK_EXPORTS
  BOOL AssumeNetAddress;
  BOOL NoWarnOnError;
  BOOL ClearHistoryOnStartup;
  BOOL WrapHistory;
  BOOL NoTabMicroComplete;
  BOOL NewestHistoryItemsOnTop;
  BOOL ContextMenuAutoPopup;
  BOOL ContextMenuAboveBox;
  BOOL ContextMenuExecute;
  BOOL UnixHistory;
  BOOL ExplorePaths;
  BOOL AutoComplete;
  BOOL CommaDelimiter;
  BOOL RPNCalculator;
  BOOL ClockDisappears;
	BOOL NoClearOnCommand;
  BOOL HideOnCommand;
  BOOL HideOnEscape;
  BOOL ClearOnHide;
  BOOL WinAmpDisappears;
  BOOL TabFileComplete;


  char *CommandChars[34];
/*           11111           1111122           222222           223333 */ 
/* 012345678901234           5678901           234567           890123 */ 
/* !"#$%&'()*+,-./   [0-9]   :;<=>?@   [A-Z]   [\]^_`   [a-z]   {|}~  */ 
#define COMMAND_CHAR_INDEX(C) ( ((C)<='@') ? ((C)<='/'?(C)-' ':(C)-':'+15) : ((C)<='`'?(C)-'['+22:(C)-'{'+28) )
#define IS_COMMAND_CHAR(C) ( ((C)>='!'&&(C)<='/') || ((C)>=':'&&(C)<='@') || ((C)>='['&&(C)<='`') || ((C)>='{'&&(C)<=128) )

#endif // LSXCOMMANDCLOCK_EXPORTS
    COLORREF BGColor;
    COLORREF TextColor;
    COLORREF BorderColor;
    //20021114
    char* OnFocusCommand;
    char* OnUnfocusCommand;
};

enum EventCode
{
  EventFocus, EventUnfocus
};

#define CM_TRIGGEREVENT (WM_USER + 1)


#ifndef LSXCOMMANDCLOCK_EXPORTS

struct  History {
    char *path;
    struct History *next, *prev;
};

/* Settings */
extern struct CommandSettings *cs;

#ifdef __cplusplus
extern "C" {
#endif

extern HMENU hSearchMenu, hAliasMenu, hHistoryMenu;

/* Execution */
extern void ExecCommand(const char *szCommand, BOOL alias);

/* History Services - history.c */
/*
extern void HistoryMoveNext(struct History **hist, int i);
extern void HistoryMovePrev(struct History **hist, int i);
extern void HistoryMoveToHead(struct History **hist);
extern void HistoryMoveToTail(struct History **hist);
extern int HistoryFindIndexOf(struct History **hist, const char *pszValue);
extern BOOL HistoryMoveToEntry(struct History **hist, const char *pszValue, BOOL tokenize);
extern struct History *HistoryAdd(struct History **hist, const char *pszValue, int *count);
extern void HistoryFree(struct History **hist);
extern void HistoryRemoveEntry(struct History **hist, int *count);
extern struct History *HistoryRemoveAll(struct History **hist, int *count);
*/

/* Calculator Services - calculator.c */
#define STATE_NONE 0
#define STATE_OPERAND 1
#define STATE_UNARYOP 2
#define STATE_OPERATOR 3
#define UNARY_NEG "(-)"
#define UNARY_DEC "d"
#define UNARY_BIN "b"
#define UNARY_HEX "h"
#define ERROR_POSTFIX_CONVERSION 1
#define ERROR_NULL_EXPRESSION    2
#define ERROR_CONVERSIONS_DONE   3
#define ERROR_EVALUATE           4

/*extern double Evaluate(char *expr, int *error, char *replacement, size_t replace_size);*/

void Complain(const char *sFormat, ...);
void ComplainExecError(int hError, const char *sLiteralCommand, const char *sExpandedCommand);

void ReadCommandChars(struct CommandSettings *pSettings);
void FreeCommandChars(struct CommandSettings *pSettings);


/* hopefully MBCS-safe string stuff - parse.c */ 
#define HASH unsigned long
BOOL Parse_GetHashedToken(const char *Source, const char **pNext,
    char *OrigCase, unsigned *pcchOrigCase,
    HASH *hash, char *KeyCase, unsigned *pcchKeyCase);
const char *Parse_SkipSpace(const char *s);
unsigned Parse_ReplaceSearchArgs(const char *URL, const char *Args, char *Buffer, unsigned cchCapacity);
unsigned Parse_ReplaceArgs(const char *Command, const char *Args, char *Buffer, unsigned cchCapacity);
char *Parse_FindLastFileToken(char *sBuffer, unsigned cchPrefix);

#define HISTORY_PAGE_FRACTION 1/3
#define FILES_PAGE_SIZE       10

/* history list - history.c */ 
void History_Load(struct CommandSettings *cs);
void History_Unload(struct CommandSettings *cs);
void History_Clear(struct CommandSettings *cs);
void History_Add(const char *line);
void History_Browse(int Direction);
const char *History_GetCurrent(void);
unsigned History_Complete(char *buffer, unsigned cchBuffer, unsigned cchPrefix);
void History_CreateMenu(struct CommandSettings *cs);

/* file completion list - history.c */ 
unsigned Files_Complete(char *sBuffer, unsigned cchBuffer, unsigned cchPrefix, int Direction);
void Files_Close(void);

/* search engine list - search.c */ 
void Search_Load(struct CommandSettings *cs);
void Search_Unload(struct CommandSettings *cs);
void Search_Execute(HWND hWnd, HASH hash, const char *key, const char *args);

/* alias list - search.c */ 
void Alias_Load(struct CommandSettings *cs);
void Alias_Unload(struct CommandSettings *cs);
BOOL Alias_Execute(HASH hash, const char *key, const char *args);

#ifdef __cplusplus
}
#endif


#else // LSXCOMMANDCLOCK_EXPORTS

/* Timer Procedures - must be declared for LSXCommandClock */
VOID CALLBACK TimerProc(HWND hWnd, UINT uMsg, UINT idEvent, DWORD dwTime);
VOID CALLBACK WinAmpTimerProc(HWND hWnd, UINT uMsg, UINT idEvent, DWORD dwTime);

#endif //LSXCOMMANDCLOCK_EXPORTS

#endif //LSCOMMAND_H